# Internet-in-a-Box System
# By Braddock Gaskill, 16 Feb 2013
from whoosh.qparser import QueryParser

from utils import whoosh2dict, whoosh_open_dir_32_or_64


class WikipediaSearch(object):
    def __init__(self, index_dir):
        """Initialize a search object.
        index_dir is the Whoosh index directory to use."""
        self.index_dir = index_dir

    def search(self, query, page=1, pagelen=20):
        """Return a sorted list of results.
        pagelen specifies the number of hits per page.
        page specifies the page of results to return (first page is 1)
        Set pagelen = None or 0 to retrieve all results.
        """
        query = unicode(query)  # Must be unicode
        ix = whoosh_open_dir_32_or_64(self.index_dir)
        with ix.searcher() as searcher:
            query = QueryParser("title", ix.schema).parse(query)
            if pagelen is not None and pagelen != 0:
                try:
                    results = searcher.search_page(query, page, pagelen=pagelen,
                                                   sortedby="score", reverse=True)
                except ValueError, e:  # Invalid page number
                    results = []
            else:
                results = searcher.search(query, limit=None,
                                          sortedby="score", reverse=True)
            #r = [x.items() for x in results]
            r = whoosh2dict(results)
        ix.close()
        return r

    def count(self, query):
        """Return total number of matching documents in index"""
        query = unicode(query)  # Must be unicode
        ix = whoosh_open_dir_32_or_64(self.index_dir)
        with ix.searcher() as searcher:
            query = QueryParser("title", ix.schema).parse(query)
            results = searcher.search(query)
            n = len(results)
        ix.close()
        return n



# ZIM file URL views (for Wikipedia)
import os
import re

from flask import Blueprint, Response, render_template, request, flash, url_for, abort
from flask.ext.babel import gettext as _
from whoosh import scoring

from zimpy import ZimFile
from config import config

from whoosh_search import paginated_search
from .endpoint_description import EndPointDescription

DEFAULT_RESULTS_PER_PAGE = 20

blueprint = Blueprint('zim_views', __name__,
                      template_folder='templates', static_folder='static')

def load_zim_file(humanReadableId):
    zim_dir = config().get_path("ZIM", "wikipedia_zim_dir")
    zim_fn = os.path.join(zim_dir, humanReadableId + ".zim")
    return ZimFile(zim_fn)

def replace_paths(top_url, html):
    replace = u"\\1\\2" + top_url + "/\\3/"
    html = re.sub(u'(href|src)(=["\']/)([A-Z\-])/', replace, html)
    html = re.sub(u'(@import[ ]+)(["\']/)([A-Z\-])/', replace, html)
    return html

def mangle_article(html, mimetype, humanReadableId):
    if mimetype in ['text/html; charset=utf-8', 'stylesheet/css', 'text/html']:
        try:
            html = html.decode('utf-8','replace')
        except UnicodeDecodeError:
            try:
                print "utf-8 decoding failed, falling back to latin1"
                html = html.decode('latin1')
            except:
                print "utf-8 and latin1 decoding failed"
                return html
        html = replace_paths("iiab/zim/" + humanReadableId, html)
    return html

@blueprint.route('/<humanReadableId>')
def zim_main_page_view(humanReadableId):
    """Returns the main page of the zim file"""
    zimfile = load_zim_file(humanReadableId)
    try:
        article, mimetype, ns = zimfile.get_main_page()
        html = mangle_article(article, mimetype, humanReadableId)
        return Response(html, mimetype=mimetype)
    except OSError as e:
        html = "<html><body>"
        html += "<p>" + _('Error accessing article.') + "</p>"
        html += "<p>" + _('Exception: %(error)s', error=str(e)) + "</p>\n"
        html += "</body></html>"
        return Response(html)

@blueprint.route('/<humanReadableId>/<namespace>/<path:url>')
def zim_view(humanReadableId, namespace, url):
    zimfile = load_zim_file(humanReadableId)
    article, mimetype, ns = zimfile.get_article_by_url(namespace, url)
    if article is None:
        abort(404)
    html = mangle_article(article, mimetype, humanReadableId)
    return Response(html, mimetype=mimetype)

@blueprint.route('/iframe/<humanReadableId>')
def iframe_main_page_view(humanReadableId):
    url = url_for('zim_views.zim_main_page_view', humanReadableId=humanReadableId)
    return render_template('zim/iframe.html', main_page=True, url=url, humanReadableId=humanReadableId)

@blueprint.route('/iframe/<humanReadableId>/<namespace>/<path:url>')
def iframe_view(humanReadableId, namespace, url):
    url = url_for('zim_views.zim_view', humanReadableId=humanReadableId, namespace=namespace, url=url)
    return render_template('zim/iframe.html', main_page=False, url=url, humanReadableId=humanReadableId)

@blueprint.route('/search/<humanReadableId>')
def search(humanReadableId):
    query = request.args.get('q', '').strip()
    pagination = None
    if query:
        index_base_dir = config().get_path("ZIM", "wikipedia_index_dir")
        index_dir = os.path.join(index_base_dir, humanReadableId)
        page = int(request.args.get('page', 1))
        # Set a higher value for the title field so it is weighted more
        (pagination, suggestion) = paginated_search(index_dir, ["title", "content"], query, page, weighting=scoring.BM25F(title_B=1.0))
    else:
        flash(_('Please input keyword(s)'), 'error')
    return render_template('zim/search.html', humanReadableId=humanReadableId, pagination=pagination, suggestion=suggestion, keywords=query, endpoint_desc=EndPointDescription('zim_views.search', {'humanReadableId':humanReadableId}))



# -*- coding: utf-8 -*-

from flask.ext.sqlalchemy import SQLAlchemy
db = SQLAlchemy()

#from flask.ext.mail import Mail
#mail = Mail()

#from flask.ext.cache import Cache
#cache = Cache()

#from flask.ext.login import LoginManager
#login_manager = LoginManager()



# Simple global configuration system
# Retrieve it with config()
import os
import platform

from ConfigParser import SafeConfigParser
from json import dumps, loads

from utils import run_mount


global_config = None


class IiabConfig(SafeConfigParser):
    """Just some conveniences for IIAB configuration"""

    def all_items(self):
        """Return a dictionary of sections with a dictionary of name/value pairs"""
        j = {}
        for section in self.sections():
            js = {}
            for name, value in self.items(section):
                js[name] = self.get(section, name)
            j[section] = js
        return j

    def all_items_to_str(self):
        return dumps(self.all_items(), indent=4)

    def __str__(self):
        return self.all_items_to_str()

    def get_json(self, section, name):
        """Load a configuration value string and interpret
           it as a JSON structure"""
        return loads(self.get(section, name))

    def verify_knowledge_dir(self, path=None):
        """Verify that path is a valid knowledge_dir directory"""
        if path is None:
            path = self.get('DEFAULT', 'knowledge_dir')
        return os.path.isdir(path)

    def get_knowledge_dir(self):
        """Find a knowledge dataset directory.  Return None if
        the configured knowledge_dir setting did not exist
        and a search for the knowledge dir either was not done
        or failed, depending on the search_for_knowledge_dir setting"""
        do_search = self.getboolean('DEFAULT', 'search_for_knowledge_dir')
        path = self.get('DEFAULT', 'knowledge_dir')
        if self.verify_knowledge_dir(path):
            return path
        if do_search:
            mount_points = run_mount()
            for mp in mount_points:
                if mp[0][0] == '/':
                    path = os.path.join(mp[1], 'knowledge')
                    if self.verify_knowledge_dir(path):
                        self.set('DEFAULT', 'knowledge_dir', path)
                        return path
            # Check current directory
            if self.verify_knowledge_dir('knowledge'):
                self.set('DEFAULT', 'knowledge_dir', 'knowledge')
                return 'knowledge'
        return None

    def get_path(self, section, name):
        """Used to get a path relative to the knowledge/ dir.
        This will first search for the knowledge/ dir on mounted volumes if
        search_for_knowledge_dir is true"""
        self.get_knowledge_dir()
        return self.get(section, name)

    def get_default(self, section, name, default):
        """Get a value if it exists, otherwise return default"""
        if not self.has_option(section, name):
            return default
        return self.get(section, name)


def load_config(config_files=[]):
    """First load iiab/defaults.ini, which is required.
    Then load optional files in the config_files array
    if they exist."""
    global global_config
    package_dir = os.path.dirname(__file__)
    master_config_file = os.path.join(package_dir, 'defaults.ini')
    config = IiabConfig()
    config.readfp(open(master_config_file, 'r'))
    config.read(config_files)
    # Set the arch variable to our cpu architecture
    config.set('DEFAULT', 'arch', platform.machine())
    global_config = config
    return global_config


def config():
    """Return the global ConfigParser object.
    If the config files have not yet been read, than this
    will load them"""
    global global_config
    if global_config is None:
        raise Exception("config() called before load_config()")
    return global_config



# -*- coding: utf-8 -*-
"""
    This module provides a simple WSGI profiler middleware for finding
    bottlenecks in web application.

    Example usage::

        from timepro_flask import TimeProMiddleware
        app = TimeProMiddleware(app)
"""
import sys
import time
import os.path

import timepro


class MergeStream(object):
    """An object that redirects `write` calls to multiple streams.
    Use this to log to both `sys.stdout` and a file::

        f = open('profiler.log', 'w')
        stream = MergeStream(sys.stdout, f)
        profiler = ProfilerMiddleware(app, stream)
    """

    def __init__(self, *streams):
        if not streams:
            raise TypeError('at least one stream must be given')
        self.streams = streams

    def write(self, data):
        for stream in self.streams:
            stream.write(data)


class TimeProMiddleware(object):
    """Simple profiler middleware.  Wraps a WSGI application and profiles
    a request.  This intentionally buffers the response so that timings are
    more exact.

    By giving the `profile_dir` argument, pstat.Stats files are saved to that
    directory, one file per request. Without it, a summary is printed to
    `stream` instead.

    For the exact meaning of `sort_by` and `restrictions` consult the
    :mod:`profile` documentation.

    .. versionadded:: 0.9
       Added support for `restrictions` and `profile_dir`.

    :param app: the WSGI application to profile.
    :param stream: the stream for the profiled stats.  defaults to stderr.
    :param sort_by: a tuple of columns to sort the result by.
    :param restrictions: a tuple of profiling strictions, not used if dumping
                         to `profile_dir`.
    :param profile_dir: directory name to save pstat files
    """

    def __init__(self, app, stream=None,
                 sort_by=('time', 'calls'), restrictions=(), profile_dir=None):
        self._app = app
        self._stream = stream or sys.stdout
        self._sort_by = sort_by
        self._restrictions = restrictions
        self._profile_dir = profile_dir

    def __call__(self, environ, start_response):
        response_body = []

        def catching_start_response(status, headers, exc_info=None):
            start_response(status, headers, exc_info)
            return response_body.append

        def runapp():
            appiter = self._app(environ, catching_start_response)
            response_body.extend(appiter)
            if hasattr(appiter, 'close'):
                appiter.close()

        url = environ.get('PATH_INFO').strip('/').replace('/', '.') or 'root'
        timepro.timepro().activate()
        timepro.timepro().start(url)
        start = time.time()
        runapp()
        body = ''.join(response_body)
        elapsed = time.time() - start
        timepro.timepro().end(url)
        if elapsed > 1.0:
            timepro.timepro().log_all()
            timepro.timepro().deactivate()

        if self._profile_dir is not None:
            prof_filename = os.path.join(self._profile_dir,
                                         '%s.%s.%06dms.%d.prof' % (
                                             environ['REQUEST_METHOD'],
                                             environ.get('PATH_INFO').strip('/').replace('/', '.') or 'root',
                                             elapsed * 1000.0,
                                             time.time()
                                        ))
            #p.dump_stats(prof_filename)

        else:
            #stats = Stats(p, stream=self._stream)
            #stats.sort_stats(*self._sort_by)

            #self._stream.write('-' * 80)
            #self._stream.write('\nPATH: %r\n' % environ.get('PATH_INFO'))
            #stats.print_stats(*self._restrictions)
            #self._stream.write('-' * 80 + '\n\n')
            pass

        return [body]



import os
from glob import glob
import logging
import time

from flask import Blueprint, render_template

from config import config
from zimpy import ZimFile
from iso639 import iso6392
from kiwix import Library
import timepro
import babel.numbers

blueprint = Blueprint('wikipedia_views', __name__,
                      template_folder='templates')

logger = logging.getLogger(__name__)

#timepro.profile_and_print()
def organize_books_by_language(filenames, library_file):
    if not os.path.exists(library_file):
        logger.error("Can not find Kiwix library file: %s" % library_file)

    kiwix_lib = Library(library_file)

    languages = {}

    for zim_fn in filenames:
        zim_obj = ZimFile(zim_fn)
        t0 = time.time()
        book_data = zim_obj.metadata()
        if time.time() - t0 > 0.5:
            logger.info("Metadata search took %f sec for %s" % (time.time() - t0, zim_fn))

        # Make sure book has metadata, if not look up in kiwix library
        # This solution allows us to add zim files not in the kiwix library
        # while still having a backup for files that do not have metadata
        if not book_data.has_key('language'):
            logger.info("No metadata, looking for book in kiwix library: %s" % zim_fn)
            book_data = kiwix_lib.find_by_uuid(zim_obj.get_kiwix_uuid())
            if book_data == None:
                logger.info("Book missing in kiwix library as well, skipping: %s" % zim_fn)
                continue

        # Decode strings from UTF-8 into unicode objects
        for k,v in book_data.items():
            if type(v) is str:
                book_data[k] = v.decode('utf-8')

        # Format article count as string with commas
        articleCount = zim_obj.header['articleCount']
        #book_data['articleCount'] = "{:,d}".format(articleCount)
        book_data['articleCount'] = babel.numbers.format_number(articleCount)
        book_data['humanReadableId'] = os.path.splitext(os.path.basename(zim_fn))[0]

        if not languages.has_key(book_data['language']):
            lang_data = {}
            if iso6392.has_key(book_data['language']):
                lang_data['languageEnglish'] = iso6392[book_data['language']]['english']
            else:
                lang_data['languageEnglish'] = "Unknown: " + book_data['language']
            lang_data['languageCode'] = book_data['language']
            lang_data['books'] = []
            lang_data['articleCount'] = 0
            languages[book_data['language']] = lang_data
        else:
            lang_data = languages[book_data['language']]

        lang_data['books'].append(book_data)
        lang_data['articleCount'] = lang_data['articleCount'] + articleCount

    langs = languages.values()
    langs.sort(key=lambda x: -x['articleCount'])
    return langs

@blueprint.route('/')
def wikipedia_view():
    wikipedia_zim_dir = config().get('ZIM', 'wikipedia_zim_dir')
    library_file = config().get('ZIM', 'kiwix_library_file')
    old_library_file = config().get('ZIM', 'old_kiwix_library_file')
    # Old location before being moved, for backwards compatibility
    if not os.path.exists(library_file):
        logger.info("Kiwix library file not found at: %s, using old location: %s" % (library_file, old_library_file))
        library_file = old_library_file
    langs = organize_books_by_language(glob(os.path.join(wikipedia_zim_dir, "*.zim")), library_file)
    return render_template('wikipedia_index.html', languages=langs)



from flask import Blueprint, request, render_template, session, jsonify
from config import config
from iiab import __version__

blueprint = Blueprint('settings', __name__,
                      template_folder='templates', static_folder='static')

def set_babel(babel_inst):
    "Called from app initialization to set a variable containing the Babel instance"

    global babel_instance
    babel_instance = babel_inst

def available_locales():
    "Returns a list of locales with translations available as well as the default locale"

    avail_locales =  babel_instance.list_translations()
    lang_names = [ l.language for l in avail_locales ]
    if babel_instance.default_locale.language not in lang_names:
        avail_locales.append(babel_instance.default_locale)
    return avail_locales

def languages_dict():
    langs = {}
    for locale in available_locales():
        lang_code = locale.language
        if locale.territory != None:
            lang_code += "_" + locale.territory
        langs[lang_code] = locale.get_display_name()
    return langs

def current_locale():
    accepted_langs = languages_dict().keys()
    preferred_language = session.get("preferred_language", None)
    if preferred_language != None:
        return preferred_language
    else:
        return request.accept_languages.best_match(accepted_langs)

@blueprint.route('/')
def settings_view():
    "Render the settings page"
    return render_template('settings.html', software_version=__version__)

@blueprint.route('/languages')
def languages_json():
    "Returns a JSON container with the language code and their display name"

    return jsonify(languages_dict())

@blueprint.route('/language', methods = ['GET', 'PUT'])
def user_language():
    """Returns the language code stored in the session dict on a GET,
    sets a new language into the session on a PUT"""

    if request.form.get('language', None) != None:
        lang = request.form['language']
        if not languages_dict().has_key(lang):
            return jsonify(result="not_found")
        else:
            session['preferred_language'] = lang

    # current_locale() will return language from either
    # request header or from session variable
    preferred_lang = current_locale()
    return jsonify(language=preferred_lang)



from babel.localedata import _cache, _cache_lock, load, merge
import os
import pickle

_mydirname = os.path.dirname(__file__)


def babel_patched_load(name, merge_inherited=True):
    # Load additional languages into the babel
    # cache.  We do this so we can add Haitian Creole
    # using french as a model.
    _cache_lock.acquire()
    try:
        data = _cache.get(name)
        if not data:
            # Load inherited data
            if name == 'root' or not merge_inherited:
                data = {}
            else:
                parts = name.split('_')
                if len(parts) == 1:
                    parent = 'root'
                else:
                    parent = '_'.join(parts[:-1])
                data = load(parent).copy()
            filename = os.path.join(_mydirname, '%s.dat' % name)
            fileobj = open(filename, 'rb')
            try:
                if name != 'root' and merge_inherited:
                    merge(data, pickle.load(fileobj))
                else:
                    data = pickle.load(fileobj)
                _cache[name] = data
            finally:
                fileobj.close()
        return data
    finally:
        _cache_lock.release()



#!/usr/bin/env python

from xml.etree import ElementTree as etree
from base64 import b64decode
import iso639
import os

def clean_book(book):
    """Fixes up the book data"""
    clean_book = {}
    for k, v in book.items():
        if k == "favicon":
            v = b64decode(v)
        elif k == "mediaCount":
            v = int(v)
        elif k == "size":
            v = int(v)
        elif k == "language":
            if v == 'en':  # Mislabel, replace with 3-letter label
                v = 'eng'
            if v in iso639.iso6392:
                clean_book["languageEnglish"] = iso639.iso6392[v]['english']
            else:
                clean_book["languageEnglish"] = v
        clean_book[k] = v
    if 'language' not in clean_book:
        title = clean_book.get('title', '')
        if title.find(" ml ") != -1:
            lang = 'mal'
        elif title.find(" zh ") != -1:
            lang = 'zho'
        else:
            # Assume english
            lang = 'eng'
        clean_book['language'] = lang
        clean_book['languageEnglish'] = iso639.iso6392[lang]['english']
    return clean_book


class Library(object):
    def __init__(self, xml_filename):
        self.books = {}
        self._parse_library(xml_filename)

    def _parse_library(self, library_xml_filename):
        """Parse a kiwix library xml file"""
        with open(library_xml_filename, "r") as f:
            et = etree.parse(f)
            root = et.getroot()
            self.books = root.findall("book")
        self.books = map(clean_book, self.books)

    def find_by_uuid(self, uuid):
        for book in self.books:
            if book['id'] == uuid:
                return book
        return None

    def books_by_language(self):
        """Get a list of all unique languages found in the library,
        sorted in decreasing order of total number of articles in that language"""
        langs = dict()
        for book in library:
            lang = book['language']
            langEng = book['languageEnglish']
            articles = book['articleCount']
            self.books = []
            if lang in langs:
                articles += langs[lang].get('articleCount', 0)
                self.books = langs[lang].get('self.books', [])
            self.books.append(book)
            langs[lang] = {
                'language': lang,
                'languageEnglish': langEng,
                'articleCount': articles,
                'self.books': self.books
            }
        langs = langs.values()
        langs.sort(key=lambda x: -x['articleCount'])
        return langs



# Top level URL views
from flask import (Blueprint, make_response, render_template,
                   send_file, redirect, abort, Markup)
from flask.ext.babel import gettext as _

from config import config
from utils import mdns_resolve

blueprint = Blueprint('top_views', __name__,
                      template_folder='templates', static_folder='static')


@blueprint.route('/')
def index():
    error = None
    if config().get_knowledge_dir() is None:
        error = _('Could not find knowledge directory containing the Internet-in-a-Box dataset.  ')
        error += _("The configured '<strong>knowledge_dir</strong>' path is '<strong>%(config_dir)s</strong>' ", config_dir=config().get('DEFAULT', 'knowledge_dir'))
        error += _(" and the '<strong>search_for_knowledge_dir</strong>' setting is ")
        if config().getboolean('DEFAULT', 'search_for_knowledge_dir'):
            error += _('<strong>ON</strong>, so all mounted filesystems were checked.')
        else:
            error += _('<strong>OFF</strong>, so other mounted filesystems were NOT checked.')
    if error is None:
        return render_template("home.html")
    else:
        return render_template("home.html", error=Markup(error))


@blueprint.route('detect')
def detect_view():
    """Detect if a valid IIAB installation exists on the localhost
    or the local network (using mDNS).  Returns a 404 if no
    IIAB installation found."""
    localhost_iiab = config().get_knowledge_dir()
    if localhost_iiab is not None:
        return "true"
    try:
        address = mdns_resolve("know.local")
        if address is not None:
            return "true"
    except:
        pass
    abort(404)


@blueprint.route('redirect')
def redirect_view():
    """Redirect the browser to our index or, if the
    local machine is not configured, attempt to
    resolve the mdns name know.local"""
    localhost_iiab = config().get_knowledge_dir()
    if localhost_iiab is not None:
        return index()
    try:
        address = mdns_resolve("know.local")
        if address is not None:
            return redirect("http://" + address)
    except:
        pass
    return index()  # This will display an error, but that is better than nothing


# This is a hack because of the double //static path issue
#@blueprint.route('static/<path:filename>')
#def static(filename):
#    return blueprint.send_static_file(filename)


@blueprint.route('test')
def test():
    print "TEST"
    return make_response((send_file('/var/www/foo.webm', mimetype="video/webm"), 200, {'Accept-Ranges': 'bytes'}))



# OpenStreetMap URL views
from flask import Blueprint, Response
import os

from osmtile import TileSet, meta_load_one
from config import config

blueprint = Blueprint('map_views', __name__,
                      template_folder='templates', static_folder='static')


@blueprint.route('/tile/<int:z>/<int:x>/<int:y>.png')
def tile(z, x, y):
    path = os.path.join(config().get_path('OSM', 'openstreetmap_dir'), 'mod_tile64')
    tileset = TileSet(path, 'default', METATILE=64, flatter=True)
    tile = meta_load_one(tileset, x, y, z)
    return Response(tile, mimetype='image/png')
